#include <gl/glut.h>
#include <gl/gl.h>
#include "EXTbasetypes.h"
#include "EXTmacros.h"
#include "extoglmacros.h"

#include "quadtree.h"
#include "task.h"

#define ABS(x)		( (x) > 0 ? (x) : (-(x) ) )
#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

#define glVertex glVertex3f

ELEVType		**field;
unsigned char	**light;
float			C;

/***************************************************
*
*
*
*
***************************************************/

/********************************************************
**
**	_name_
**
**	DESCRIPTION: _
**
**
**W
**	NOTES: _
**
**
**
********************************************************/

/********************************************************
**														*
**	quadAllocateMemory ()								*
**														*
**	DESCRIPTION: Allocates memory for a QUADTREE		*
**	and returns the address of the allocated pointer.	*
**														*
**														*
**	NOTES: None.										*
**														*
**														*
**														*
********************************************************/

void *quadAllocateMemory ()
{
	QUADTREE *tree;

	tree = (QUADTREE *) malloc (sizeof (QUADTREE) );

	return tree;
}

/********************************************************
**														*
**	quadInit											*
**														*
**	DESCRIPTION: Initializes a quadtree data type. Sets	*
**	a given dimensions, min. and max. level. Returns	*
**	the address of the allocated memory.				*
**														*
**	NOTES: None											*
**														*
**														*
**														*
********************************************************/

void *quadInit (int pDimensions)
{

/****************************************************************************
*																			*
*	QUADTREE: pTree - Variable to hold the address of the new quadtree		*
*																			*
****************************************************************************/
	
	QUADTREE *tree;

	tree = quadAllocateMemory ();

	tree->dim		= pDimensions;

	return tree;
}

/********************************************************
**														*
**	quadAllocateNodeMatrix								*
**														*
**	DESCRIPTION: Allocates the 2D matrix of the			*
**	quadnodes.											*
**														*
**														*
**	NOTES: None											*
**														*
**														*
**														*
********************************************************/

void quadAllocateNodeMatrix (QUADTREE *pTree)
{

/****************************************************************************
*																			*
*	INDEXType: i - Counter													*
*																			*
****************************************************************************/
	
	register INDEXType i;

	pTree->nodeMatrix = (QUADNODE **) calloc (pTree->dim, sizeof (QUADNODE *) );
	
	for (i = 0; i < pTree->dim; i++)
		pTree->nodeMatrix [i] = (QUADNODE *) malloc (pTree->dim * sizeof (QUADNODE) );	
}	

/********************************************************
**														*
**	quadFreeMemory										*
**														*
**	DESCRIPTION: Frees the QUADTREE, including its nodes*
**														*
**														*
**														*
**	NOTES: None											*
**														*
**														*
**														*
********************************************************/

void quadFreeMemory (QUADTREE *pTree)
{

/****************************************************************************
*																			*
*	INDEXType: i - Counter													*
*																			*
****************************************************************************/

	register INDEXType i;

	for (i = 0; i < pTree->dim; i++)
		free (pTree->nodeMatrix[i]);		
	
	free (pTree->nodeMatrix);	
	free (pTree);
	
}

/********************************************************
**														*
**	quadClearTriangulation								*
**														*
**	DESCRIPTION: Resets the node matrix to 0.			*
**														*
**														*
**														*
**	NOTES: Must be called before initial traversal		*
**	of the tree.										*
**														*
**														*
********************************************************/

void quadClearTriangulation (QUADTREE *pTree)
{

/****************************************************************************
*																			*
*	INDEXType:	i - Row counter (slow)										*
*				j - Column counter (fast)									*
*																			*
****************************************************************************/

	register INDEXType i, j, end = pTree->dim;

	for (i = 0; i < end; i++)
		for (j = 0; j < end; j++)
			pTree->nodeMatrix[i][j].status = 0;
			
}

/********************************************************
**														*
**	quadPropagateRoughness								*
**														*
**	DESCRIPTION: Propogates roughness values in-order	*
**	to avoid difference of > 1 in levels of detail		*
**														*
**														*
**	NOTES: Must be called in-order to avoid cracks in	*
**	the mesh											*
**														*
**														*
********************************************************/

void quadPropagateRoughness (QUADTREE *pTree, INDEXType x, INDEXType y, INDEXType w)
{

/****************************************************************************
*																			*
*	unsigned char: max - Temporary variable to hold maximum roughness value	*
*	of adjacent nodes.														*
*																			*
****************************************************************************/
	register unsigned char max;

	if (w != 1)
	{
		w >>= 1;

		quadPropagateRoughness (pTree, x - w, y - w, w);
		quadPropagateRoughness (pTree, x - w, y + w, w);
		quadPropagateRoughness (pTree, x + w, y + w, w);
		quadPropagateRoughness (pTree, x + w, y - w, w);

		w <<= 1;
	}

	max = pTree->nodeMatrix[x][y].roughness;
	if (x - w - w >= 0)
		max = MAX (max, pTree->nodeMatrix[x - w - w][y].roughness);
	if (x + w + w < pTree->dim)
		max = MAX (max, pTree->nodeMatrix[x + w + w][y].roughness);
	if (y + w + w < pTree->dim)
		max = MAX (max, pTree->nodeMatrix[x][y + w + w].roughness);
	if (y - w - w >= 0)
		max = MAX (max, pTree->nodeMatrix[x][y - w - w].roughness);
	
	pTree->nodeMatrix[x][y].roughness = max;
}

/********************************************************
**														*
**	quadGetRoughness									*
**														*
**	DESCRIPTION: Calculates surface roughness, measured	*
**	by the offset of the middle-points from the average	*
**	of the two adjacet corners.							*
**														*
**	NOTES: None.										*
**														*
**														*
**														*
********************************************************/

float quadGetRoughness (HEIGHTFIELD *pField, int x, int y, int w)
{
	
/****************************************************************************
*																			*
*	float:		edges[6] - stores average height of 6 edges					*
*																			*
*	ELEVType:	accum - Sum of: | MIDDLE-POINT - (CORNER1 + CORNER2) |		*
*								|			   -------------------	 |		*
*								|						2			 |		*
*																			*
****************************************************************************/

	float	edges[6];
	float	max;

	edges [0] = pField->elev[x - w][y + w] - pField->elev[x - w][y - w];
	edges [1] = pField->elev[x + w][y + w] - pField->elev[x - w][y + w];
	edges [2] = pField->elev[x + w][y - w] - pField->elev[x + w][y + w];
	edges [3] = pField->elev[x - w][y - w] - pField->elev[x + w][y - w];
	edges [4] = pField->elev[x + w][y + w] - pField->elev[x - w][y - w];
	edges [5] = pField->elev[x - w][y + w] - pField->elev[x + w][y - w];

	max = ABS(pField->elev[x - w][y + w] + (edges[0] / 2.0) - pField->elev[x - w][y] );
	max = MAX(max, fabs(pField->elev[x + w][y + w] + (edges[1] / 2.0) - pField->elev[x][y + w] ) );
	max = MAX(max, fabs(pField->elev[x + w][y - w] + (edges[2] / 2.0) - pField->elev[x + w][y] ) );
	max = MAX(max, fabs(pField->elev[x - w][y - w] + (edges[3] / 2.0) - pField->elev[x][y - w] ) );
	max = MAX(max, fabs(pField->elev[x + w][y + w] + (edges[4] / 2.0) - pField->elev[x][y] ) );
	max = MAX(max, fabs(pField->elev[x - w][y + w] + (edges[5] / 2.0) - pField->elev[x][y] ) );

	return max;
}

/********************************************************
**														*
**	quadProcessNode										*
**														*
**	DESCRIPTION: Processes each node, and sets its pre-	*
**	liminary roughness value. The roughness value is	*
**	derieved from the greatest child-nodes' roughness	*
**	value. For leaf-nodes, the value is calculated using*
**	quadGetRoughness ()									*
**														*
**	NOTES: max_rougness is a variable that used to 		*
**	clamp the roughness values to [0, 255]. It can be	*
**	tweaked in-order to multiply the roughness values.	*
**														*
********************************************************/

void quadProcessNode(QUADTREE *pTree, HEIGHTFIELD *pField, INDEXType x, INDEXType y, INDEXType w)
{

/****************************************************************************
*																			*
*	ELEVType: roughness - Temporary variable to hold roughness value		*
*																			*
*	unsigned char: max - Temporary variable to hold maximum roughness value	*
*																			*
****************************************************************************/
	
	register ELEVType roughness;
	register unsigned char max;

	if (w != 1)
	{
		w >>= 1;

		quadProcessNode (pTree, pField, x - w, y - w, w);
		quadProcessNode (pTree, pField, x - w, y + w, w);
		quadProcessNode (pTree, pField, x + w, y + w, w);
		quadProcessNode (pTree, pField, x + w, y - w, w);

		max = MAX(pTree->nodeMatrix[x - w][y - w].roughness, pTree->nodeMatrix[x - w][y + w].roughness);
		max = MAX(max, pTree->nodeMatrix[x + w][y + w].roughness);
		max = MAX(max, pTree->nodeMatrix[x + w][y - w].roughness);
		pTree->nodeMatrix[x][y].roughness = max;
	}
	else
	{
		roughness = pTree->mul_roughness * quadGetRoughness(pField, x, y, w);

		if (roughness > 255)
			pTree->nodeMatrix[x][y].roughness = 255;
		else
			pTree->nodeMatrix[x][y].roughness = roughness;
	}
}

/********************************************************
**														*
**	quadProcessTree										*
**														*
**	DESCRIPTION: Processes the tree - sets the			*
**	roughness values of each node, and propogates the	*
**	values upwards.										*
**														*
**	NOTES: None.										*
**														*
**														*
**														*
********************************************************/

void quadProcessTree (QUADTREE *pTree, HEIGHTFIELD *pField)
{
	pTree->center_and_width = ( (pTree->dim -  1) / 2);

	taskNew ("Allocating memory for matrix");
	quadAllocateNodeMatrix (pTree);
	taskOk ();	

	pTree->mul_roughness = 5.0;
	taskNew ("Calculating surface roughess");
	quadProcessNode (pTree, pField, pTree->center_and_width, pTree->center_and_width, pTree->center_and_width);
	taskOk ();

	taskNew ("Propogating roughness upwards");
	quadPropagateRoughness (pTree, pTree->center_and_width, pTree->center_and_width, pTree->center_and_width);
	quadClearTriangulation (pTree);
	taskOk ();

}

__inline void quadPickColorUB (const unsigned char lightness, const ELEVType elev)
{
	register float lumina;
	lumina = lightness / 255.0f;

	if (elev < 90)
	{
		glColor3f (1.0 * lumina, 0.8 * lumina, 0.2 * lumina);
		return;
	}
	if (elev < 175)
	{
		glColor3f (0.54 * lumina, 0.8 * lumina, 0.18 * lumina);
		return;
	}
	if (elev < 181)
	{
		glColor3f (0.69 * lumina, 0.7 * lumina, 0.71 * lumina);
		return;
	}
	glColor3ub (lightness, lightness, lightness);
	return;
}

__inline void quadPickColorF (const float lightness, const ELEVType elev)
{
	if (elev < 90)
	{
		glColor3f (1.0 * lightness, 0.8 * lightness, 0.2 * lightness);
		return;
	}
	if (elev < 175)
	{
		glColor3f (0.54 * lightness, 0.8 * lightness, 0.18 * lightness);
		return;
	}
	if (elev < 181)
	{
		glColor3f (0.69 * lightness, 0.7 * lightness, 0.71 * lightness);
		return;
	}
	
	glColor3f (lightness, lightness, lightness);
	return;
}

void quadRenderNode (QUADTREE *pTree, const HEIGHTFIELD *pField, const INDEXType x, const INDEXType y, INDEXType w, const float *corners)
{

/****************************************************
*													*
*	INDEXType:	x_?_w, y_?_w - x - w, y - w, etc.	*
*				w2 =  w								*
*					 ---							*
*					  2								*
*													*
*	unsigned char:	lightnessc - Lightness value	*
*					(corner vertices)				*
*													*
*	float:	dc - Two-dimensional array used for		*
*	calculating shared-edge corners					*
*			lightness - Lightness value (middle		*
*			vertices)								*
*													*
*	macros: stat	- gives status of current node	*
*			stm		- status matrix					*
*			field	- height-field					*
*			blend	- calculates elevation/lighting	*
*													*
*			(z1 * i) + (0.5 * (z2 * z3)(255 - i)	*
*	using:	------------------------------------	*
*							255						*
*													*
****************************************************/

#define stat pTree->nodeMatrix[x][y].status
#define stm(p, l) (pTree->nodeMatrix[p][l].status)
#define blend(z1, z2, z3, i) ( ( ( (float) z1 * (float) i) + ( 0.5f * (255.0f - (float) i) * ( (float) z2 + (float) z3 ) ) ) / 255.0f)
#define lblend(z1, z2, z3, i) ( ( ( (float) z1 * (float) i) + ( 0.5f * (255.0f - (float) i) * ( (float) z2 + (float) z3 ) ) ) / 65280.0f)

	register INDEXType x_m_w, x_p_w, y_m_w, y_p_w;
	register INDEXType w2 = w >> 1;
	register float lightness;
	register unsigned char lightnessc;
	float dc[4][4];	

	if (stat > 1) /* If should draw node */
	{
		x_m_w = x - w;				/* Calculate ?_?_w's */
		y_m_w = y - w;
		x_p_w = x + w;
		y_p_w = y + w;		
		
		pTree->nTris++;

		glBegin (GL_TRIANGLE_FAN);	/* Begin new triangle fan */
		{
			lightness = lblend(light[x][y], ( (float) light[x_m_w][y_m_w] + (float) light[x_p_w][y_m_w]) / 2.0f, ( (float) light[x_p_w][y_p_w] + (float) light[x_m_w][y_p_w]) / 2.0f, stm(x,y) );
			quadPickColorF (lightness, field[x][y]);

			glTexCoord2f (x, y);	/* Draw vetrex 0 */
			glVertex (x, y, blend(field[x][y], ( (float) field[x_m_w][y_m_w] + (float) field[x_p_w][y_m_w]) / 2.0f, ( (float) field[x_p_w][y_p_w] + (float) field[x_m_w][y_p_w]) / 2.0f, stm(x,y) ) );
			
			lightnessc = light[x_m_w][y_m_w];
			quadPickColorUB (lightnessc, field[x_m_w][y_m_w]);
			glTexCoord2f (x_m_w, y_m_w);	/* Draw vertex 1 */
			glVertex (x_m_w, y_m_w, corners[0]);
			
			if (x_m_w - w > -1)	/* If in range of [0, DIM] */
				if (pTree->nodeMatrix[x_m_w - w][y].status > 0)	/* If node has been visited or drawn */
				{
					lightness = lblend(light[x_m_w][y], light[x_m_w][y_p_w], light[x_m_w][y_m_w], stm(x_m_w ,y) );
					quadPickColorF (lightness, field[x_m_w][y]);
					glTexCoord2f (x_m_w, y);	/* Draw vertex 2 */
					glVertex (x_m_w, y, blend(field[x_m_w][y], field[x_m_w][y_p_w], field[x_m_w][y_m_w], stm(x_m_w ,y) ) );
				}
			
			lightnessc = light[x_m_w][y_p_w];
			quadPickColorUB (lightnessc, field[x_m_w][y_p_w]);
			glTexCoord2f (x_m_w, y_p_w);	/* Draw vertex 3 */
			glVertex (x_m_w, y_p_w, corners[1]);
			
			if (y_p_w + w < pTree->dim)
				if (pTree->nodeMatrix[x][y_p_w + w].status > 0)
				{
					lightness = lblend(light[x][y_p_w], light[x_m_w][y_p_w], light[x_p_w][y_p_w], stm(x, y_p_w) );
					quadPickColorF (lightness, field[x][y_p_w]);
					glTexCoord2f (x, y_p_w);	/* Draw vertex 4 */
					glVertex (x, y_p_w, blend(field[x][y_p_w], field[x_m_w][y_p_w], field[x_p_w][y_p_w], stm(x, y_p_w) ) );
				}
			
			lightnessc = light[x_p_w][y_p_w];
			quadPickColorUB (lightnessc, field[x_p_w][y_p_w]);
			glTexCoord2f (x_p_w, y_p_w);	/* Draw vertex 5 */
			glVertex (x_p_w, y_p_w, corners[2]);

			if (x_p_w + w < pTree->dim)
				if (pTree->nodeMatrix[x_p_w + w][y].status > 0)
				{
					lightness = lblend(light[x_p_w][y], light[x_p_w][y_p_w], light[x_p_w][y_m_w], stm(x_p_w ,y) );
					quadPickColorF (lightness, field[x_p_w][y]);
					glTexCoord2f (x_p_w, y);	/* Draw vertex 6 */
					glVertex (x_p_w, y, blend(field[x_p_w][y], field[x_p_w][y_p_w], field[x_p_w][y_m_w], stm(x_p_w ,y) ) );
				}
			
			lightnessc = light[x_p_w][y_m_w];
			quadPickColorUB (lightnessc, field[x_p_w][y_m_w]);
			glTexCoord2f (x_p_w, y_m_w);	/* Draw vertex 7 */
			glVertex (x_p_w, y_m_w, corners[3]);
			
			if (y_m_w - w > -1)
				if (pTree->nodeMatrix[x][y_m_w - w].status > 0)
				{
					lightness = lblend(light[x][y_m_w], light[x_m_w][y_m_w], light[x_p_w][y_m_w], stm(x, y_m_w) );
					quadPickColorF (lightness, field[x][y_m_w]);
					glTexCoord2f (x, y_m_w);	/* Draw vertex 8 */
					glVertex (x, y_m_w, blend(field[x][y_m_w], field[x_m_w][y_m_w], field[x_p_w][y_m_w], stm(x, y_m_w) ) );			
				}
			
			lightnessc = light[x_m_w][y_m_w];
			quadPickColorUB (lightnessc, field[x_m_w][y_m_w]);
			glTexCoord2f (x_m_w, y_m_w);	/* Draw vertex 1 (complete fan) */
			glVertex (x_m_w, y_m_w, corners[0]);
			
		} glEnd();		
		return;
	}
	
	/* else - refine node */

	if (stat == 0)
		return;
	
	if ( (stm(x - w2, y - w2) > 1) || (stm(x - w2, y + w2) > 1) || (stm(x + w2, y + w2) > 1) || (stm(x + w2, y - w2) > 1) ) /* If node surely have children that should be drawn */	
	{
		x_m_w = x - w;	/* Calculate ?_?_w's */
		y_m_w = y - w;
		x_p_w = x + w;
		y_p_w = y + w;

		dc[0][0] = field[x_m_w][y_m_w];
		dc[0][1] = field[x_m_w][y];
		dc[0][2] = field[x][y];
		dc[0][3] = field[x][y_m_w];

		dc[1][0] = field[x_m_w][y];
		dc[1][1] = field[x_m_w][y_p_w];
		dc[1][2] = field[x][y_p_w];
		dc[1][3] = field[x][y];

		dc[2][0] = field[x][y];
		dc[2][1] = field[x][y_p_w];
		dc[2][2] = field[x_p_w][y_p_w];
		dc[2][3] = field[x_p_w][y];

		dc[3][0] = field[x][y_m_w];
		dc[3][1] = field[x][y];
		dc[3][2] = field[x_p_w][y];
		dc[3][3] = field[x_p_w][y_m_w];
		
		if (x_m_w - w > -1)	/* Check for shared edges at lower LOD. */
		{
			if (stm(x_m_w - w, y) > 1)	/* If exists - modify corner values to avoid cracks */
			{
				dc[0][1] = blend(field[x_m_w][y], field[x_m_w][y_m_w], field[x_m_w][y_p_w], stm(x_m_w, y) );
				dc[1][0] = dc[0][1];
			}
		}

		if (x_p_w + w < pTree->dim)
		{
			if (stm(x_p_w + w, y) > 1)
			{
				dc[2][3] = blend(field[x_p_w][y], field[x_p_w][y_p_w], field[x_p_w][y_m_w], stm(x_p_w, y) );
				dc[3][2] = dc[2][3];
			}
		}

		if (y_p_w + w < pTree->dim)
		{
			if (stm(x, y_p_w + w) > 1)
			{
				dc[1][2] = blend(field[x][y_p_w], field[x_p_w][y_p_w], field[x_m_w][y_p_w], stm(x, y_p_w) );
				dc[2][1] = dc[1][2];
			}
		}

		if (y_m_w - w > -1)
		{
			if (stm(x, y_m_w - w) > 1)
			{
				dc[0][3] = blend(field[x][y_m_w], field[x_p_w][y_m_w], field[x_m_w][y_m_w], stm(x, y_m_w) );
				dc[3][0] = dc[0][3];
			}
		}
	}

	quadRenderNode (pTree, pField, x - w2, y - w2, w2, dc[0]);	/* Call children */
	quadRenderNode (pTree, pField, x - w2, y + w2, w2, dc[1]);
	quadRenderNode (pTree, pField, x + w2, y + w2, w2, dc[2]);
	quadRenderNode (pTree, pField, x + w2, y - w2, w2, dc[3]);
}
__inline int quadClip (const EXTcamera *pCamera, const INDEXType x_l, const INDEXType x_r, const INDEXType y_d, const INDEXType y_u)
{
	register float f;
	
	f = (pCamera->ax1 * x_l) + (pCamera->by1 * y_d) + pCamera->d1;
	if (f > 0)
	{
		f = (pCamera->ax1 * x_r) + (pCamera->by1 * y_d) + pCamera->d1;
		if (f > 0)
		{
			f = (pCamera->ax1 * x_l) + (pCamera->by1 * y_u) + pCamera->d1;
			if (f > 0)
			{
				f = (pCamera->ax1 * x_r) + (pCamera->by1 * y_u) + pCamera->d1;
				if (f > 0)
					return 1;
			}
		}
	}

	f = (pCamera->ax2 * x_l) + (pCamera->by2 * y_d) + pCamera->d2;	
	if (f < 0)
	{
		f = (pCamera->ax2 * x_r) + (pCamera->by2 * y_d) + pCamera->d2;
		if (f < 0)
		{
			f = (pCamera->ax2 * x_l) + (pCamera->by2 * y_u) + pCamera->d2;
			if (f < 0)
			{
				f = (pCamera->ax2 * x_r) + (pCamera->by2 * y_u) + pCamera->d2;
				if (f < 0)
					return 1;
			}
		}
	}


	return 0;
}

void quadClearNode	(const QUADTREE *pTree, INDEXType x, INDEXType y, INDEXType w)
{
	pTree->nodeMatrix[x - w][y - w].status = 0;
	pTree->nodeMatrix[x - w][y + w].status = 0;
	pTree->nodeMatrix[x + w][y + w].status = 0;
	pTree->nodeMatrix[x + w][y - w].status = 0;	
}


/********************************************************
**														*
**	quadTriangulateNode									*
**														*
**	DESCRIPTION: Creates the triangulation for current	*
**	iteration. Notice that distance is calculated by	*
**	L1norm, which creates the familiar diamond-shaped	*
**	iso-detailed areas on the mesh.						*
**	In order to avoid cracks and corrupted tringulati-	*
**	ons, each leaf sets it actual children value to 0.	*
**	This enables the program to avoid quadClearTring.	*
**	for each iteration.									*
**														*
**	NOTES: None.										*
**														*
**														*
********************************************************/

void quadTriangulateNode (const QUADTREE *pTree, const HEIGHTFIELD *pField, const INDEXType x, const INDEXType y, INDEXType w, const EXTcamera *pCamera, int level)
{

/****************************************************************************
*																			*
*	float:	f		- Stores the result of equation (5)						*
*			k		- Stores the roughness value							*
*			dist	- Distances between eye and vertex 0					*
*																			*
*			Equation (5):													*
*																			*
*							  d												*
*					f =	---------------			(subdivide if (1 > f) )		*
*						lC * (rougness)										*
*																			*
****************************************************************************/
	
	register float f, k = pTree->nodeMatrix[x][y].roughness;
	register float dist = ABS(pCamera->location.x - (float) x) + ABS(pCamera->location.y - (float) y) + ABS(pCamera->location.z - (float) pField->elev[x][y]);
	register unsigned char fc;

	if (quadClip (pCamera, x - w, x + w, y - w, y + w) )
	{
		pTree->nodeMatrix[x][y].status = 0;		
		return;
	}	

	f = dist / ( (float) w * C * k);	
		
	if (w == 1)
		if (f < 1)
			f = 1;
		
	if (f >= 1)
	{
		if (f > 2)
			f = 2;

		f = 2.0 - f;

		f = 2.0 + (253.0 * f);

		fc = f;

		pTree->nodeMatrix[x][y].status = f;
		pTree->nodeMatrix[x - w][y].status = MAX(pTree->nodeMatrix[x - w][y].status, fc);
		pTree->nodeMatrix[x + w][y].status = MAX(pTree->nodeMatrix[x + w][y].status, fc);
		pTree->nodeMatrix[x][y - w].status = MAX(pTree->nodeMatrix[x][y - w].status, fc);
		pTree->nodeMatrix[x][y + w].status = MAX(pTree->nodeMatrix[x][y + w].status, fc);		
		
		if (w > 1)
		{
			w >>= 1;

			pTree->nodeMatrix[x - w][y - w].status = 0;
			pTree->nodeMatrix[x - w][y + w].status = 0;
			pTree->nodeMatrix[x + w][y + w].status = 0;
			pTree->nodeMatrix[x + w][y - w].status = 0;

		}
		return;
	}		
	
	pTree->nodeMatrix[x][y].status = 1;

	w >>= 1;
	level++;
		
	quadTriangulateNode (pTree, pField, x - w, y - w, w, pCamera, level);
	quadTriangulateNode (pTree, pField, x - w, y + w, w, pCamera, level);
	quadTriangulateNode (pTree, pField, x + w, y + w, w, pCamera, level);
	quadTriangulateNode (pTree, pField, x + w, y - w, w, pCamera, level);

	return;
}

/********************************************************
**														*
**	quadSetCameraClipBox								*
**														*
**	DESCRIPTION: Calculates the lines used to clip		*
**	againt the frustum.									*
**														*
**	NOTES: None											*
**														*
**														*
**														*
********************************************************/

void quadSetCameraClipbox (EXTcamera *pCamera)
{
	EXTpoint3 u;
	double lambda;

	lambda = pCamera->half_fov;// + (fabs(sin(fabs(pCamera->roll) * EXT_PI / 90.0) ) * 0.4);

	pCamera->ax2 = sin (-pCamera->angle + lambda);
	pCamera->by2 = cos (-pCamera->angle + lambda);
	
	pCamera->ax1 = sin (-pCamera->angle - lambda);
	pCamera->by1 = cos (-pCamera->angle - lambda);

	if (pCamera->pitch < 0)
	{	
		u.x = pCamera->location.x - (-cos(pCamera->angle) * sin (pCamera->pitch) * (pCamera->location.z + 30.0) );
		u.y = pCamera->location.y - (-sin(pCamera->angle) * sin (pCamera->pitch) * (pCamera->location.z + 30.0) );
	}
	else
	{
		u.x = pCamera->location.x - (cos(pCamera->angle) * sin (pCamera->pitch) * 60.0);
		u.y = pCamera->location.y - (sin(pCamera->angle) * sin (pCamera->pitch) * 60.0);
	}	
	pCamera->d1 = - ( (pCamera->ax1 * u.x) + (pCamera->by1 * u.y) );
	pCamera->d2 = - ( (pCamera->ax2 * u.x) + (pCamera->by2 * u.y) );	
}

void quadRenderTree(QUADTREE *pTree, HEIGHTFIELD *pField, EXTcamera *pCamera, float pC)
{
	float corners[4] = { 0, 0, 0, 0 };

	C = pC;	

	light = pField->lightmap;
	field = pField->elev;	

	pTree->nTris = 0;

	quadSetCameraClipbox(pCamera);
	quadTriangulateNode (pTree, pField, pTree->center_and_width, pTree->center_and_width, pTree->center_and_width, pCamera, 0);
	quadRenderNode (pTree, pField, pTree->center_and_width, pTree->center_and_width, pTree->center_and_width, corners);	
}