#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "heightfield.h"
#include "EXTbasetypes.h"
#include "EXTmacros.h"

#define MAX(a, b) ((a) > (b) ? (a) : (b))

void *hfAllocateMemory (int pWidth, int pHeight)
{
	HEIGHTFIELD *field;
	int i;

	field = (HEIGHTFIELD *) malloc (sizeof(HEIGHTFIELD) );
	field->elev		= (ELEVType **) calloc (pWidth, sizeof (ELEVType	*) );
	field->lightmap = (unsigned char **) calloc (pWidth, sizeof (unsigned char *) );
	
	for (i = 0; i < pWidth; i++)
	{
		field->elev[i]	= (ELEVType *) calloc (pHeight, sizeof (ELEVType) );
		field->lightmap [i] = (unsigned char *) calloc (pHeight, sizeof (unsigned char) );
	}
		
	field->dim_w = pWidth;
	field->dim_h = pHeight;
	return field;
}

void hfFreeMemory (HEIGHTFIELD *pField)
{
	int i;

	for (i = 0; i < pField->dim_w; i++)
	{
		free (pField->elev[i]);
		free (pField->lightmap [i]);
	}
	
	free (pField);
}

void hfResetField (HEIGHTFIELD *pField)
{
	int i, j;

	for (i = 0; i < pField->dim_w; i++)
		for (j = 0; j < pField->dim_h; j++)
			pField->elev[i][j] = 0;
}

void hfRetrieveField (HEIGHTFIELD *pField, char *pDatabase)
{
	FILE *data_stream;
	int i;

	data_stream = fopen (pDatabase, "rb");

	if (! data_stream)
	{
		printf ("\nDatabase (%s) not found. Resetting height-field.\n", pDatabase);
		hfResetField (pField);
	}
	else
		for (i = 0; i < pField->dim_w; i++)
			fread (pField->elev [i], sizeof (ELEVType), pField->dim_h, data_stream);

}

void hfTwo (HEIGHTFIELD *pField)
{
	double p = 6.28 / (0.5 * (double) pField->dim_w);
	INDEXType i, j;

	for (i = 0; i < pField->dim_w; i++)
		for (j = 0; j < pField->dim_h; j++)
			pField->elev[i][j] = (double) 4.0 * cos (p * (double) i) * sin (p * (double) j);
}

void hfSmoothField (HEIGHTFIELD *pField, int iterations)
{
	int i, j, k;

	for (k = 0; k < iterations; k++)
		for (i = 1;i < pField->dim_w - 1; i++)
			for (j = 1; j < pField->dim_h - 1; j++)
				pField->elev[i][j] = (pField->elev[i][j] + pField->elev[i][j + 1] + pField->elev[i + 1][j]) / 3.0;
}

float	hfGetElevationf (HEIGHTFIELD *pField, EXTpoint3f *location)
{
	register float Dx, Dy;
	register float P1, P2, P3, P4;
	register INDEXType i, j;

	i = location->x;
	j = location->y;

	P1 = pField->elev [i][j];
	P2 = pField->elev [i + 1][j];
	P3 = pField->elev [i][j + 1];
	P4 = pField->elev [i + 1][j + 1];
	Dx = location->x - (float) i;
	Dy = location->y - (float) j;
	
	return (Dx * Dy * P1) + ( (1.0f - Dx) * Dy * P2 ) + ( Dx * (1.0f - Dy) * P3) + ( (1.0f - Dx) * (1.0f - Dy) * P4);
}


void hfLightUpField (HEIGHTFIELD *pField)
{

/****************************************************************************
*																			*
*	EXTvector3: normal		- surface normal								*
*				edges[2]	- array to hold	vectors for cross-product		*
*																			*
*	float:		dot			- stores the value of the dot product			*
*																			*
****************************************************************************/

	EXTvector3 sun_vector;

	EXTvector3 normal, edges[2];
	register float dot;
	INDEXType i, j, k;
	
	sun_vector.x = 0.33035042472811;
	sun_vector.y = 0.0;
	sun_vector.z = 0.94385835636602;

	for (i = 1; i < pField->dim_w; i++)
		for (j = 1; j < pField->dim_h; j++)
		{
			edges[0].x = -2;
			edges[0].y = 0;
			edges[0].z = pField->elev[i - 1][j] - pField->elev[i][j];

			edges[1].x = 0;
			edges[1].y = -2;
			edges[1].z = pField->elev[i][j - 1] - pField->elev[i][j];

			EXT_VECTOR_CROSS_PRODUCT (edges[0], edges[1], normal);

			EXT_VECTOR_NORMALIZE (normal);				

			dot = EXT_VECTOR_DOT_PRODUCT (sun_vector, normal);

			if (dot < 0)
				dot = 0;
			pField->lightmap[i][j] = 55.0 + (dot * 205.0);
		}
	
	for (k = 0; k < 3; k++)
		for (i = 0; i < pField->dim_w - 2; i += 1)
			for (j = 0; j < pField->dim_h - 2; j += 1)
				pField->lightmap[i][j] = (pField->lightmap[i][j] + pField->lightmap[i + 1][j + 1] + pField->lightmap[i + 1][j] + pField->lightmap[i][j + 1]) / 4.0;			
}

void hfCalculateHistogramUB (HEIGHTFIELD *pField)
{
	unsigned long count [20], max, k;
	int i, j;
	char legend[10];

	printf ("\nCalculating histogram... ");

	for (i = 0; i < 20; i++)
		count [i] = 0;

	for (i = 0; i < pField->dim_w; i++)
		for (j = 0; j < pField->dim_h; j++)
			count [pField->elev [i][j] / 13] += 1;

	max = 0;
	for (i = 0; i < 20; i++)
		max = MAX(max, count [i]);

	if (max == 0)
		return;

	printf ("Done.");
	
	for (i = 0; i < 20; i++)
	{
		printf ("\n");
		for (j = 0; j < 8; j++)
			legend [j] = ' ';

		sprintf (legend, "%d-%d", i * 13, (i + 1) * 13 - 1);
		for (j = 0; j < 8; j++)
			printf ("%c", legend [j]);

		printf (":");		
		for (k = 0; k < count [i] * 50 / max; k++)
			printf ("|");
	}

	printf ("\n");
}